import { world, ItemStack, system, EquipmentSlot, MolangVariableMap } from "@minecraft/server"
import { ActionFormData, MessageFormData, ModalFormData } from "@minecraft/server-ui"
import { TerminalSimulator } from "./terminal"

const cableId = "effect99:storage_link_cable"
const iconBox = ['box_yellow', 'box_green', 'box_blue', 'box_pink', 'box_orange', 'box_gray', 'box_withe']
const equipmentSlots = [EquipmentSlot.Head, EquipmentSlot.Chest, EquipmentSlot.Legs, EquipmentSlot.Feet, EquipmentSlot.Offhand]

system.beforeEvents.watchdogTerminate.subscribe(s => { s.cancel = true; })

const worldTicks = () => system.runTimeout(() => {
    for (const player of world.getAllPlayers()) {
        const locateContainer = player.getTags().find(f => { return f.includes("Container #") })
        if (locateContainer) {
            const item = player.getComponent('inventory').container.getItem(player.selectedSlot)
            if (!item || item.typeId !== 'effect99:remote_terminal_controller') {
                player.removeTag(locateContainer)
                player.onScreenDisplay.setActionBar(`§cDeleting location`)
                worldTicks()
                return
            }
            const { x, y, z } = player.location
            const coords = locateContainer.split("[")[1].replace("]", "").split(",")
            const dx = parseInt(coords[0])
            const dy = parseInt(coords[1])
            const dz = parseInt(coords[2])

            const distance = Math.sqrt(((x - 0.5) - dx) ** 2 + ((y - 0.5) - dy) ** 2 + ((z - 0.5) - dz) ** 2).toFixed(2)
            try { player.dimension.spawnParticle(`effect:binary`, { x: dx + Math.random(), y: dy + Math.random(), z: dz + Math.random() }, new MolangVariableMap()) } catch (e) { }
            player.onScreenDisplay.setActionBar(`§hLocating: §6${locateContainer.slice(0, locateContainer.indexOf("[") - 1)}\n§7Block distance: §f${distance}`)
        }
    }
    worldTicks()
}, 5)
worldTicks()

system.afterEvents.scriptEventReceive.subscribe(e => {
    const block = e.sourceBlock
    if (block && block.typeId === cableId) {
        const directions = { "down": block.below(), "up": block.above(), "north": block.north(), "south": block.south(), "west": block.west(), "east": block.east() }

        for (const [direction, blockAtDirection] of Object.entries(directions)) {
            if (validContainers.some(f => { return blockAtDirection.typeId.includes(f) })) {
                const state = `cable:${direction}`
                if (!block.permutation.getState(state)) {
                    block.setPermutation(block.permutation.withState(state, 1))
                }
            }
        }
    }
    else if (e.id === "terminal:remote") {
        remoteTerminalController(e)
    }
})

world.afterEvents.itemStartUseOn.subscribe(i => {
    let item = i.itemStack
    if (item && item.typeId === 'effect99:remote_terminal_controller' && i.block.typeId === 'effect99:storage_manager_terminal') {
        addTerminalToController(i, item)
    }
    else if (i.block.typeId === 'effect99:storage_manager_terminal' && !i.block.permutation.getState("terminal:state") && (!item || !item.typeId.includes(cableId))) {
        screenAnimation(i, 1)
        terminalOnSound(i)
        terminal(i)
    }
})

function remoteTerminalController(e) {
    const player = e.sourceEntity
    const dimension = e.sourceEntity.dimension
    const inventory = player.getComponent("inventory").container
    const controller = inventory.getItem(player.selectedSlot)
    const remoteTerminal = new ActionFormData().title(`Remote Terminal Controller`)
    const terminalList = controller.getLore()
    const pingList = []

    for (const terminals of terminalList) {
        const coords = terminals.split("[")[1].replace("]", "").split(",")
        const x = parseInt(coords[0])
        const y = parseInt(coords[1])
        const z = parseInt(coords[2])

        let signal = ""
        const block = dimension.getBlock({ x, y, z })
        if (block && block.typeId === `effect99:storage_manager_terminal`) {
            let ping = 0
            if (dimension.getBlock({ x: x + 48, y, z })) ping++
            if (dimension.getBlock({ x: x - 48, y, z })) ping++
            if (dimension.getBlock({ x, y, z: z + 48 })) ping++
            if (dimension.getBlock({ x, y, z: z - 48 })) ping++

            if (ping <= 1) signal = ""
            else if (ping >= 2 && ping <= 3) signal = ""
            else if (ping === 4) signal = ""
        }
        pingList.push(signal)
        remoteTerminal.button(`§8< §7Terminal ${signal} §8> \n${terminals}`, 'textures/ui/more_icon')
    }
    remoteTerminal.body(`§8Terminals: ${terminalList.length}/20\n${player.getTags().some(f => { return f.includes("Container #") }) ? `§7Locating: §9${player.getTags().find(f => { return f.includes("Container #") })} \n§8Remove the device from your hand to erase the location.` : ``}`)
    if (!controller.getLore()[0]) remoteTerminal.button(`§8< §7There are no saved terminals §8>`, 'textures/ui/more_icon')

    remoteTerminal.show(player).then(r => {
        if (!r.canceled && controller.getLore()[0]) {
            const noSignal = pingList[r.selection] === ""
            const terminalManage = new ActionFormData().title(`Remote Terminal Controller`).button(noSignal ? ` No signal ` : `Connect to terminal`, noSignal ? `textures/ui/noSignal` : null)
                .button(`§8< §7Remove Terminal §8>\n Delete controller terminal`, `textures/ui/remove_terminal`)

            if (noSignal) terminalManage.body(`§8There is no signal, it is likely that you are very far from the terminal or that terminal no longer exists.`)

            terminalManage.show(player).then(t => {
                if (!t.canceled) {
                    if (t.selection === 0) {
                        if (noSignal) { remoteTerminalController(e); return }
                        player.onScreenDisplay.setActionBar(`${pingList[r.selection]} Connecting to the terminal...`)
                        const coords = terminalList[r.selection].split("[")[1].replace("]", "").split(",")
                        const x = parseInt(coords[0])
                        const y = parseInt(coords[1])
                        const z = parseInt(coords[2])
                        const block = e.sourceEntity.dimension.getBlock({ x, y, z })

                        if (block) {
                            terminalOnSound({ source: e.sourceEntity, block })
                            terminal({ source: e.sourceEntity, block })
                            dimension.spawnParticle("signal_emitter", { x, y: y + 5, z }, new MolangVariableMap())
                        }
                    }
                    else if (t.selection === 1) {
                        controller.setLore(terminalList.filter(f => { return f !== terminalList[r.selection] }))
                        inventory.setItem(player.selectedSlot, controller)
                        remoteTerminalController(e)
                    }
                } else remoteTerminalController(e)
            })
        }
    })
}

function addTerminalToController(i, item) {
    const inventory = i.source.getComponent("inventory").container
    const remoteTerminal = new ActionFormData().title(`Remote Terminal Controller`)
    const terminalList = item.getLore()
    const { x, y, z } = i.block

    let controller
    for (const location of terminalList) {
        const coords = location.split("[")[1].replace("]", "").split(",")
        const tx = parseInt(coords[0])
        const ty = parseInt(coords[1])
        const tz = parseInt(coords[2])
        if (x === tx && y === ty && z === tz) controller = location
    }
    remoteTerminal.button(`§8< ${controller ? `§7Terminal  §8>\n${controller}` : terminalList.length >= 20 ? `§7You can only save a maximum of 20 terminals per controller §8>` : `§7Add terminal to controller §8>`}`, controller ? `textures/ui/main_menu_icon` : terminalList.length >= 20 ? null : `textures/ui/add_terminal`)
    if (controller) remoteTerminal.button(`§8< §7Remove Terminal §8>\n Delete controller terminal`, `textures/ui/remove_terminal`)

    remoteTerminal.show(i.source).then(t => {
        item = inventory.getItem(i.source.selectedSlot)
        if (!t.canceled && item.typeId === 'effect99:remote_terminal_controller') {
            if (t.selection === 0) {
                if (!controller && terminalList.length < 20) {
                    const saveTerminal = new ModalFormData().title(`Save terminal as`).textField("\n§8Name:", "", `Terminal #${terminalList.length + 1}`).toggle("Cancel")
                    saveTerminal.show(i.source).then(r => {
                        item = inventory.getItem(i.source.selectedSlot)
                        if (!r.canceled && !r.formValues[1] && item.typeId === 'effect99:remote_terminal_controller') {
                            if (r.formValues[0].length) {
                                const tName = r.formValues[0]

                                terminalList.push(`§r§f${tName.length > 24 ? tName.slice(0, 21) + "..." : tName} §9[${[x, y, z]}]`)
                                item.setLore(terminalList)
                                inventory.setItem(i.source.selectedSlot, item)
                                addTerminalToController(i, item)
                            }
                        }
                    })
                } else if (controller) {
                    screenAnimation(i, 1)
                    terminalOnSound(i)
                    terminal(i)
                }
            }
            else if (t.selection === 1) {
                item.setLore(terminalList.filter(f => { return f !== controller }))
                inventory.setItem(i.source.selectedSlot, item)
            }
        }
    })
}

function itemData(item, enchantment, durability, remark) {
    return `${remark ? "§c" : `${enchantment ? "§5" : "§7"}`}${item.typeId.split(":")[1].slice(0, 1).toUpperCase() + item.typeId.split(":")[1].slice(1).replace(/_/g, " ")}${remark ? ` ` : ``} \n§r§8${durability ? `§9Durability: ${durability.maxDurability - durability.damage}/${durability.maxDurability}` : `${enchantment ? `§9${enchantment.type.id.slice(0, 1).toUpperCase() + enchantment.type.id.slice(1).replace(/_/g, " ")} ${enchantment.level}` : ""}`}${item.isStackable ? `Amount: ${item.amount}` : ``}  ${unknownInfo.has(item.typeId) ? `Unknown   ` : ''}`
}
function screenAnimation(i, state = 0) {
    const { x, y, z } = i.block
    if (!i.source.dimension.getBlock({ x, y, z })) { i.source.onScreenDisplay.setActionBar("§c No signal "); return }
    i.block.setPermutation(i.block.permutation.withState("terminal:state", state))
}
function terminalOffSound(i) {
    const { x, y, z } = i.block
    world.playSound("terminal.old_pc_off", { x, y, z })
}
function terminalOnSound(i) {
    const { x, y, z } = i.block
    world.playSound("terminal.old_pc_boot", { x, y, z })
    world.playSound("terminal.old_pc_sound", { x, y, z })
}

async function terminal(i, search, loadedChests, page = 0, searches) {
    const { x, y, z } = i.block
    const dimension = i.block.dimension
    const storage = new ActionFormData().title(`Storage Management Terminal`).button(`<     >`, 'textures/ui/more_icon')
    const settings = i.source.getTags().find(f => { return f.includes('{"terminalSettings":[') })
    const [, settB, settC, , , , settG] = settings ? JSON.parse(settings).terminalSettings : [2, 100, false, false, false, false, false]

    screenAnimation(i, 1)
    let chests = loadedChests ? loadedChests : await allChests(x, y, z, dimension)
    const loadeds = loadedChests ? loadedChests : chests
    if (loadeds.length) storage.button(`§hSearch items\n${search ? `§6Results of: ` + search : '§8Chest list:'}`, 'textures/ui/search_icon')

    const chestsWithItems = search ? searches ? searches : loadeds.filter(({ x, y, z }) => {
        const block = dimension.getBlock({ x, y, z })
        if (!block || !block.getComponent("inventory")) return false
        const container = block.getComponent("inventory").container
        const notEmpty = container.size != container.emptySlotsCount
        if (!notEmpty) return false

        let found = false
        if (search) {
            for (let i = 0; i < container.size; i++) {
                const slot = container.getItem(i)
                if (!slot || !(settG ? slot.typeId.split(":")[1] : slot.typeId).includes(search)) continue
                found = true
                break
            }
        } else found = true
        return notEmpty && found
    }) : settC ? loadeds.filter(({ x, y, z }) => {
        const block = dimension.getBlock({ x, y, z })
        if (!block || !block.getComponent("inventory")) return false
        const container = block.getComponent("inventory").container
        return container.size != container.emptySlotsCount
    }) : loadeds

    let maxPerPage = settB
    const startIndex = page * maxPerPage
    const endIndex = startIndex + maxPerPage
    const maxChests = chestsWithItems.slice(startIndex, endIndex)
    const thereIsNextPage = chestsWithItems.slice(startIndex, endIndex + 1).length > maxPerPage
    const totalPages = Math.ceil(chestsWithItems.length / maxPerPage)
    let foundChests = []
    let chestCount = 0

    if (search) {
        for (const { x, y, z } of maxChests) {
            const block = dimension.getBlock({ x, y, z })
            if (!block || !block.getComponent("inventory")) continue
            let containerName = block.typeId.includes(`chest`) ? `Chest` : block.typeId.includes(`_shulker_box`) ? `Shulker Box` : `Barrel`
            const container = block.getComponent("inventory").container

            for (let i = 0; i < container.size; i++) {
                if (container.size === container.emptySlotsCount) break
                const slot = container.getItem(i)
                if (!slot) continue
                if ((settG ? slot.typeId.split(":")[1] : slot.typeId).includes(search)) {
                    chestCount++
                    storage.button(`${containerName} §8#${startIndex + chestCount}${search ? ` ` : ''}\n§7Slots: ${container.size - container.emptySlotsCount} §8- §9${x}, ${y}, ${z}`, `textures/ui/${container.size > 27 ? `chests_icon` : block.typeId.includes(`chest`) ? `chest_icon` : block.typeId.includes(`_shulker_box`) ? `shulker_icon` : `barrel_icon`}`)
                    foundChests.push({ x, y, z })
                    break
                }
            }
        }
        chests = foundChests
    }
    else for (const { x, y, z } of maxChests) {
        const chestBlock = dimension.getBlock({ x, y, z })
        if (!chestBlock || !chestBlock.getComponent("inventory")) continue
        const containerName = chestBlock.typeId.includes(`chest`) ? `Chest` : chestBlock.typeId.includes(`_shulker_box`) ? `Shulker Box` : `Barrel`
        const container = chestBlock.getComponent("inventory").container
        const isEmpty = container.size - container.emptySlotsCount
        chestCount++
        storage.button(`${containerName} §8#${startIndex + chestCount}${search ? ` ` : ''}\n${isEmpty ? `§7Slots: ${isEmpty}` : `§8Empty`} §8- §9${x}, ${y}, ${z}`, `textures/ui/${container.size > 27 ? `chests_icon` : chestBlock.typeId.includes(`chest`) ? `chest_icon` : chestBlock.typeId.includes(`_shulker_box`) ? `shulker_icon` : `barrel_icon`}`)
    }

    if (thereIsNextPage) storage.button(`Next page`, 'textures/ui/next_icon')
    if (page) storage.button(`Previous page`, 'textures/ui/back_icon')

    storage.body(`§8Chests: ${thereIsNextPage || page || search ? `${chestCount}/${chestsWithItems.length}  ${totalPages > 1 ? `Page: ${page + 1}/${totalPages}` : ''}` : chestCount} ${search && chestCount ? `\n§7Chests with the result: ${search}` : search ? `\n§cNo results for: ${search}` : ``}`)
    storage.show(i.source).then(r => {
        if (!r.canceled) {
            if (r.selection === 1) {
                const searchMenu = new ModalFormData().title(`Search`).textField("", "Search").toggle("Delete searches and cancel")
                screenAnimation(i, 1)
                searchMenu.show(i.source).then(r => {
                    if (!r.canceled) {
                        if (!r.formValues[1]) { terminal(i, r.formValues[0].trim().toLowerCase(), loadeds, 0); world.playSound("terminal.pc_pip", { x, y, z }) }
                        else terminal(i, 0, loadeds, 0)
                    }
                    else terminal(i, search, loadeds, page, searches)
                    screenAnimation(i)
                })
            }
            else if (r.selection === 0) {
                moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount)
            }
            else if (r.selection >= 2 && r.selection <= chestCount + 1) {
                getChestsItems(i, { i: r.selection - 2, n: startIndex }, [search ? chests[r.selection - 2] : maxChests[r.selection - 2]], dimension, 0, loadeds, search, page, chestsWithItems)
            }
            else if (r.selection === chestCount + 2) {
                if (thereIsNextPage) terminal(i, search, loadeds, page + 1, chestsWithItems)
                else if (page) terminal(i, search, loadeds, page - 1, chestsWithItems)
            }
            else if (r.selection === chestCount + 3) terminal(i, search, loadeds, page - 1, chestsWithItems)
        } else if (search) terminal(i, 0, loadeds, 0); else terminalOffSound(i)
    })
    screenAnimation(i)
}

function moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount, empty, alerts) {
    const options = new ActionFormData().title(`More options`).body(`§8§¨F:\\Users\\${i.source.name}\\Terminal\\Options>`)
        .button(`All containers${search && chestCount ? `\n§6Chests with the result: ` + search : search ? `\n§cNo results for: ${search}` : ``}`, `textures/ui/all_chest_icon`).button(`§hSave items\n ${alerts === 3 ? `§aItem saved successfully.` : `§8Open inventory`}`, 'textures/ui/save_icon')
        .button(`§hQuick save\n${empty ? `§cThere are no items in your hand` : alerts === 4 ? `§aItem saved successfully.` : `§8The item in your hand.`}`, 'textures/ui/quick_save_icon').button(`§hSave inventory\n${alerts === 1 ? `§6Some items have been saved` : alerts === 2 ? `§aItems saved successfully.` : ``}`, 'textures/ui/inventory').button(`Terminal Settings`, `textures/ui/settings_icon`)
        .button(`Terminal commands`, `textures/ui/terminal_icon`)

    options.show(i.source).then(r => {
        if (!r.canceled) {
            if (r.selection === 0) {
                if (!chestCount) { moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount); return }
                getChestsItems(i, 0, null, dimension, 0, loadeds, search, page, search ? chestsWithItems : null)
            }
            else if (r.selection === 1) {
                getInventory(i, null, { search, loadeds, page, dimension, chestsWithItems, chestCount })
            }
            else if (r.selection === 2) {
                if (i.source.getComponent('inventory').container.getItem(i.source.selectedSlot)) {
                    saveItems(i, { type: 1, slot: i.source.selectedSlot }, null, true, { search, loadeds, page, dimension, chestsWithItems, chestCount })
                } else moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount, true)
            }
            else if (r.selection === 3) {
                const saveAll = new ModalFormData().title(`Save Inventory`).toggle(`Save inventory`, true).toggle(`Save equipment`).toggle(`Save hotbar`)
                    .dropdown(`Save in`, [`First valid containers`, `Last valid containers`, `Containers with similar items`, `Random container`]).toggle("Cancel")

                saveAll.show(i.source).then(n => {
                    if (!n.canceled && !n.formValues[4]) {
                        saveInventory(i, null, n.formValues[0], n.formValues[1], n.formValues[2], n.formValues[3], { search, loadeds, page, dimension, chestsWithItems, chestCount })
                    }
                    else moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount)
                })
            }

            else if (r.selection === 4) {
                const settings = i.source.getTags().find(f => { return f.includes('{"terminalSettings":[') })
                const [settA, settB, settC, settD, settE, settF, settG] = settings ? JSON.parse(settings).terminalSettings : [2, 100, false, false, false, false, false]
                const settMenu = new ModalFormData().title(`${i.source.name} Terminal Settings`).toggle("Reset settings")
                    .slider(`§cDisplaying too many buttons on the screen may cause the game to temporarily freeze.\n§8<AllChest> 54x2 full chests === 108 Buttons\n§8<AllChest>§7Max Chest Per Page`, 2, 12, 1, settA)
                    .slider(`§8<MainMenu> 1 Chest === 1 Button\n<MainMenu>§7Max Chests Per Page`, 30, 300, 1, settB)
                    .toggle(`Hide empty chests`, settC).toggle(`Only show search results`, settD).toggle(`Close menu when taking an item`, settE).toggle(`Close menu when saving an item`, settF)
                    .toggle(`Ignore identifiers when searching for Items`, settG)

                settMenu.show(i.source).then(s => {
                    if (!s.canceled) {
                        if (settings) i.source.removeTag(settings)
                        if (!s.formValues[0]) i.source.addTag(JSON.stringify({ terminalSettings: [s.formValues[1], s.formValues[2], s.formValues[3], s.formValues[4], s.formValues[5], s.formValues[6], s.formValues[7]] }))
                        moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount)
                    }
                    else moreOptions(i, search, loadeds, page, dimension, chestsWithItems, chestCount)
                })
            }
            else if (r.selection === 5) {
                terminalCommands(i, { search, loadeds, page, dimension, chestsWithItems, chestCount })
            }
        }
        else terminal(i, search, loadeds, page, chestsWithItems)
    })
}

function saveInventory(i, chest, inv, equip, hotb, savIn, p) {
    const inventory = i.source.getComponent('inventory').container
    const equipment = i.source.getComponent('equippable')

    const lastChest = savIn === 1
    const similarChest = savIn === 2
    const randomChest = savIn === 3
    let fullChests = true

    for (const { x, y, z } of chest ? chest : lastChest ? p.loadeds.reverse() : randomChest ? p.loadeds.slice().sort(() => Math.random() - 0.5) : p.loadeds) {
        const block = p.dimension.getBlock({ x, y, z })
        if (!block.getComponent("inventory")) continue
        const container = block.getComponent("inventory").container
        if (!container.emptySlotsCount) continue
        let empty = true

        if (equip && empty) for (const slot of equipmentSlots) {
            const item = equipment.getEquipmentSlot(slot).getItem()
            if (!item) continue
            else if (extractItem(item, container, slot, true)) { empty = false; break }
        }
        if (inv && empty) for (let slot = 9; slot < inventory.size; slot++) {
            const item = inventory.getItem(slot)
            if (!item) continue
            else if (extractItem(item, container, slot)) { empty = false; break }
        }
        if (hotb && empty) for (let slot = 0; slot < 9; slot++) {
            const item = inventory.getItem(slot)
            if (!item) continue
            else if (extractItem(item, container, slot)) { empty = false; break }
        }
        if (!empty) continue
    }
    function extractItem(item, container, slot, equip) {
        const durability = item.hasComponent("durability")
        const enchantments = item.hasComponent("enchantable")
        const keywords = item.typeId.slice(item.typeId.indexOf(":") + 1).split("_")
        let matched = false

        if (similarChest) {
            for (let i = 0; i < container.size; i++) {
                if (container.size === container.emptySlotsCount) break
                const matchItem = container.getItem(i)
                if (!matchItem) continue
                const matchDur = matchItem.hasComponent("durability")
                const matchEnc = matchItem.hasComponent("enchantable")

                if (matchItem.typeId === item.typeId) matched = true
                else if (matchDur && durability) matched = true
                else if (matchEnc && enchantments) matched = true
                else if (matchItem.typeId.slice(matchItem.typeId.indexOf(":") + 1).split("_").some(a => { return keywords.some(b => { return a === b }) })) matched = true
                else if (matchItem.typeId.includes("stone") && !matchItem.typeId.includes("redstone") && item.typeId.includes("stone")) matched = true
            }
        }
        if (!similarChest || matched) {
            if (container.emptySlotsCount) {
                container.addItem(item)
                equip ? equipment.setEquipment(slot, undefined) : inventory.setItem(slot, undefined)
                fullChests = false
            }
        }
        return !container.emptySlotsCount
    }
    if (fullChests) {
        const messageForm = new MessageFormData().title("Save Inventory").body({ translate: similarChest ? `terminal.messageForm.notSimilar.text` : `terminal.messageForm.fullChest${chest ? `` : `s`}.text` }).button1('Close the menu')
            .button2({ translate: `terminal.messageForm.back.chest${chest ? `` : `s`}.text` })
        messageForm.show(i.source).then(m => {
            if (!m.canceled) {
                if (m.selection === 1) chest ? getChestsItems(i, p.iChest, chest, p.dimension, p.page, p.allChestList, p.search, 0, p.searches, p.internalSearch) :
                    moreOptions(i, p.search, p.loadeds, p.page, p.dimension, p.chestsWithItems, p.chestCount); else if (m.selection === 0) terminalOffSound(i)
            }
        })
    }
    else {
        let successful = 2
        if (equip) for (const slot of equipmentSlots) {
            const item = equipment.getEquipmentSlot(slot).getItem()
            if (item) successful = 1
        }
        if (inv) for (let slot = 9; slot < inventory.size; slot++) {
            const item = inventory.getItem(slot)
            if (item) successful = 1
        }
        if (hotb) for (let slot = 0; slot < 9; slot++) {
            const item = inventory.getItem(slot)
            if (item) successful = 1
        }
        chest ? getChestsItems(i, p.iChest, chest, p.dimension, p.page, p.allChestList, p.search, 0, p.searches, p.internalSearch, successful) :
            moreOptions(i, p.search, p.loadeds, p.page, p.dimension, p.chestsWithItems, p.chestCount, false, successful)
    }
}

function getInventory(i, chest, p, otherPlayer) {
    const invMenu = new ActionFormData().title(`Inventory`).body(`§8§¨F:\\Users\\${otherPlayer ? otherPlayer.name : i.source.name}\\Terminal\\${chest ? `Containers` : `Options`}\\Inventory>`)
    const inventory = otherPlayer ? otherPlayer.getComponent('inventory').container : i.source.getComponent('inventory').container
    const itemSlot = []
    let slotCount = 0

    const equipment = otherPlayer ? otherPlayer.getComponent('equippable') : i.source.getComponent('equippable')
    for (const slot of equipmentSlots) {
        const item = equipment.getEquipmentSlot(slot).getItem()
        if (!item) continue
        const enchantment = item.hasComponent("enchantable") ? item.getComponent("enchantable").getEnchantments()[0] : null
        const durability = item.getComponent("durability")

        invMenu.button(itemData(item, enchantment, durability), 'textures/ui/equip_icon')
        itemSlot.push({ type: 0, slot: slot })
        slotCount++
    }
    for (let i = 0; i < inventory.size; i++) {
        const item = inventory.getItem(i)
        if (!item) continue
        const enchantment = item.hasComponent("enchantable") ? item.getComponent("enchantable").getEnchantments()[0] : null
        const durability = item.getComponent("durability")

        invMenu.button(itemData(item, enchantment, durability), 'textures/ui/items_icon')
        itemSlot.push({ type: 1, slot: i })
        slotCount++
    }
    invMenu.button(`Main menu`, 'textures/ui/main_menu_icon').show(i.source).then(e => {
        if (!e.canceled) {
            if (e.selection < slotCount) {
                saveItems(i, itemSlot[e.selection], chest, false, p, otherPlayer)
            }
            else if (e.selection === slotCount) {
                terminal(i, p.search, chest ? p.allChestList : p.loadeds, p.page, p.chestsWithItems)
            }
        }
        else chest ? getChestsItems(i, p.iChest, chest, p.dimension, p.page, p.allChestList, p.search, 0, p.searches, p.internalSearch) :
            moreOptions(i, p.search, p.loadeds, p.page, p.dimension, p.chestsWithItems, p.chestCount)
    })
}

function saveItems(i, items, chest, quickSave, p, otherPlayer) {
    const settings = i.source.getTags().find(f => { return f.includes('{"terminalSettings":[') })
    const [, , , , , settF,] = settings ? JSON.parse(settings).terminalSettings : [2, 100, false, false, false, false, false]
    const inventory = otherPlayer ? otherPlayer.getComponent('inventory').container : i.source.getComponent('inventory').container
    const equipment = otherPlayer ? otherPlayer.getComponent('equippable') : i.source.getComponent('equippable')
    const item = items.type ? inventory.getItem(items.slot) : equipment.getEquipmentSlot(items.slot).getItem()
    if (!item) { back(); return }
    const enchantments = item.hasComponent("enchantable") ? item.getComponent("enchantable").getEnchantments() : []
    const durability = item.getComponent("durability")
    const enchant = []

    if (enchantments.length) {
        for (const enchantment of enchantments) {
            enchant.push(` ${enchantment.type.id.slice(0, 1).toUpperCase() + enchantment.type.id.slice(1).replace(/_/g, " ")}: ${enchantment.level}\n`)
        }
    }
    const unknownItem = `§hCould not get enough information about what type of §c${item.typeId.replace(/_/g, " ")}§h this item is.`
    const itemName = `${item.typeId.split(":")[1].slice(0, 1).toUpperCase() + item.typeId.split(":")[1].slice(1).replace(/_/g, " ")}`
    const itemData = `§hID: ${item.typeId}\n${durability ? `§7Durability: ${durability.maxDurability - durability.damage}/${durability.maxDurability}\n` : ""}${item.isStackable ? `§hAmount: ${item.amount}\n` : ""}\n${enchantments.length ? `§hEnchantments:§9\n${enchant.join().replace(/,/g, "")}\n` : ``}${unknownInfo.has(item.typeId) ? `${unknownItem}\n\n` : ''}§rSave`
    const save = new ModalFormData().title(itemName)
        .slider(itemData, 1, item.amount, 1)
    if (!chest) save.dropdown(`Save in`, [`First valid container`, `Last valid container`, `Containers with similar items`, `Random container`])
    save.toggle(`Cancel`)

    save.show(i.source).then(r => {
        if (!r.canceled && !r.formValues[!chest ? 2 : 1]) {
            if (!item) { back(); return }
            const lastChest = !chest && r.formValues[1] === 1
            const similarChest = !chest && r.formValues[1] === 2
            const randomChest = !chest && r.formValues[1] === 3
            const subtracted = item.amount - r.formValues[0]
            const keywords = item.typeId.slice(item.typeId.indexOf(":") + 1).split("_")

            const durability = item.hasComponent("durability")
            const enchantments = item.hasComponent("enchantable")

            let foundSlot = false

            for (const { x, y, z } of chest ? chest : lastChest ? p.loadeds.reverse() : randomChest ? p.loadeds.slice().sort(() => Math.random() - 0.5) : p.loadeds) {
                const block = p.dimension.getBlock({ x, y, z })
                if (!block || !block.getComponent("inventory")) continue
                const container = block.getComponent("inventory").container
                if (!container.emptySlotsCount) continue

                let matched = false
                if (similarChest) {
                    for (let ind = 0; ind < container.size; ind++) {
                        if (container.size === container.emptySlotsCount) break
                        const matchItem = container.getItem(ind)
                        if (!matchItem) continue
                        const matchDur = matchItem.hasComponent("durability")
                        const matchEnc = matchItem.hasComponent("enchantable")

                        if (matchItem.typeId === item.typeId) matched = true
                        else if (matchDur && durability) matched = true
                        else if (matchEnc && enchantments) matched = true
                        else if (matchItem.typeId.slice(matchItem.typeId.indexOf(":") + 1).split("_").some(a => { return keywords.some(b => { return a === b }) })) matched = true
                        else if (matchItem.typeId.includes("stone") && item.typeId.includes("stone")) matched = true
                    }
                }
                if (!similarChest || matched) {
                    if (container.emptySlotsCount) {
                        const newStack = item.clone()
                        newStack.amount = r.formValues[0]

                        container.addItem(newStack)
                        if (subtracted) item.amount = subtracted
                        items.type ? inventory.setItem(items.slot, subtracted < 1 ? undefined : item) : equipment.setEquipment(items.slot, subtracted < 1 ? undefined : item)
                        foundSlot = true
                        break
                    }
                }
                if (foundSlot) break
            }
            if (!foundSlot) {
                const messageForm = new MessageFormData().title(itemName).body({ translate: similarChest ? `terminal.messageForm.notSimilar.text` : `terminal.messageForm.fullChest${chest ? '' : 's'}.text` }).button1('Close the menu')
                    .button2({ translate: `terminal.messageForm.back.chest${chest ? '' : 's'}.text` })
                messageForm.show(i.source).then(m => { if (!m.canceled) { if (m.selection === 1) back(); else if (m.selection === 0) terminalOffSound(i) } })
            }
            else if (!settF) back(quickSave ? 4 : 3); else terminalOffSound(i)
        }
        else back()
    })
    function back(alerts) {
        chest ? getChestsItems(i, p.iChest, chest, p.dimension, p.page, p.allChestList, p.search, 0, p.searches, p.internalSearch, alerts) :
            moreOptions(i, p.search, p.loadeds, p.page, p.dimension, p.chestsWithItems, p.chestCount, 0, alerts)
    }
}

function getChestsItems(i, iChest = { i: 0, n: 0 }, chest, dimension, page, allChestList, search, chestsPage = 0, searches, internalSearch, alerts) {
    const settings = i.source.getTags().find(f => { return f.includes('{"terminalSettings":[') })
    const [settA, , , settD, , , settG] = settings ? JSON.parse(settings).terminalSettings : [2, 100, false, false, false, false, false]
    const chests = new ActionFormData().title(!chest ? `All items` : `Container #${iChest.i + iChest.n + 1}`)
    chests.button(chest ? `§8< §hContainer options §8>\n${alerts === 1 ? `§6Some items have been saved` : alerts >= 2 && alerts <= 4 ? `§aItems saved successfully.` : alerts === 5 ? `§aItem added to your inventory` : ``}` : `§7Search \n${alerts === 5 ? `§aItem added to your inventory` : ``}`, `textures/ui/${chest ? `chest_options_icon` : `search_icon`}`)
    let slotCount = 0

    const chestsWithItems = searches ? searches : allChestList.filter(({ x, y, z }) => {
        if (chest) return false
        const block = dimension.getBlock({ x, y, z })
        if (!block || !block.getComponent("inventory")) return false
        const container = block.getComponent("inventory").container
        return container.size != container.emptySlotsCount
    })

    let maxPerPage = settA
    const startIndex = page * maxPerPage
    const endIndex = startIndex + maxPerPage
    const maxChests = chestsWithItems.slice(startIndex, endIndex)
    const thereIsNextPage = !chest && chestsWithItems.slice(startIndex, endIndex + 1).length > maxPerPage
    const totalPages = Math.ceil(chestsWithItems.length / maxPerPage)
    const itemLocation = []
    const player = i.source

    for (const { x, y, z } of !chest ? maxChests : chest) {
        const block = dimension.getBlock({ x, y, z })
        const container = block.getComponent("inventory").container

        for (let i = 0; i < container.size; i++) {
            const slot = container.getItem(i)
            if (!slot) continue
            if (settD && search && !(settG ? slot.typeId.split(":")[1] : slot.typeId).includes(search)) continue
            const remark = (search || internalSearch) && (settG ? slot.typeId.split(":")[1] : slot.typeId).includes(internalSearch ? internalSearch : search)
            const enchantment = slot.hasComponent("enchantable") ? slot.getComponent("enchantable").getEnchantments()[0] : null
            const durability = slot.getComponent("durability")

            chests.button(itemData(slot, enchantment, durability, remark), `textures/ui/${iconBox[Math.floor(Math.random() * iconBox.length)]}`)

            itemLocation.push({ x, y, z, slot: i })
            slotCount++
        }
    }
    let allSlots = 0
    for (const { x, y, z } of !chest ? chestsWithItems : []) {
        const block = dimension.getBlock({ x, y, z })
        if (!block || !block.getComponent("inventory")) continue
        const container = block.getComponent("inventory").container
        allSlots += container.size - container.emptySlotsCount
    }
    if (!chest) {
        if (thereIsNextPage) chests.button(`Next page`, 'textures/ui/next_icon')
        if (page) chests.button(`Previous page`, 'textures/ui/back_icon')
    }
    const searchResults = `\n${(search || internalSearch) ? `${internalSearch && search ? "§8" : "§7"}${search ? `Chest${chest ? `` : `s`} with the result: ` : ``}${search ? search : `Searching: ${internalSearch}`}${internalSearch && search ? `\n§7Temporary search: ` + internalSearch : ''}` : ``}`
    chests.button(`Main menu`, 'textures/ui/main_menu_icon')
    chests.body(`§8Slots: ${thereIsNextPage || page ? `${slotCount}/${allSlots}  ${totalPages > 1 ? `Page: ${page + 1}/${totalPages}` : ''}` : slotCount}${chest ? `` : `  Chests: ${Math.min(settA, chestsWithItems.length)}/${chestsWithItems.length}`} ${searchResults}`)
    chests.show(player).then(r => {
        if (!r.canceled) {
            if (r.selection === 0) {
                if (!chest) searchItems()
                else {
                    function chestOptionsMenu(empty, emptyHand) {
                        const inventory = player.getComponent('inventory').container
                        const remoteContr = inventory.getItem(player.selectedSlot) && inventory.getItem(player.selectedSlot).typeId === `effect99:remote_terminal_controller`
                        const { x, y, z } = chest[0]
                        const chestOptions = new ActionFormData().title(`Container options`).button(`§7Search \n${empty ? `§cThere are no items to search` : ``}`, 'textures/ui/search_icon').button(`§hSave items\n§8Open inventory`, 'textures/ui/save_icon')
                            .button(`§hQuick save\n${emptyHand ? `§cThere are no items in your hand` : `§8The item in your hand.`}`, 'textures/ui/quick_save_icon')
                            .button(`§hSave inventory`, 'textures/ui/inventory')
                            .button(`§hTake all the items\n${empty ? `§cThere is no item` : ``}`, 'textures/ui/items_icon')
                            .button(`§hLocate container\n${remoteContr ? `§9[${[x, y, z]}] ` : `§cYou need the remote terminal!`}`, 'textures/ui/chest_icon')
                        chestOptions.show(player).then(c => {
                            if (!c.canceled) {
                                if (c.selection === 0) {
                                    if (slotCount) {
                                        searchItems()
                                    } else chestOptionsMenu(true)
                                }
                                else if (c.selection === 1) {
                                    getInventory(i, chest, { iChest, dimension, page, allChestList, search, searches, internalSearch })
                                }
                                else if (c.selection === 2) {
                                    if (inventory.getItem(player.selectedSlot)) {
                                        saveItems(i, { type: 1, slot: player.selectedSlot }, chest, true, { iChest, dimension, page, allChestList, search, searches, internalSearch })
                                    }
                                    else chestOptionsMenu(0, true)
                                }
                                else if (c.selection === 3) {
                                    const saveAll = new ModalFormData().title(`Save Inventory`).toggle(`Save inventory`, true).toggle(`Save equipment`).toggle(`Save hotbar`).toggle("Cancel")
                                    saveAll.show(player).then(n => {
                                        if (!n.canceled && !n.formValues[3]) {
                                            saveInventory(i, chest, n.formValues[0], n.formValues[1], n.formValues[2], 0, { iChest, dimension, page, allChestList, search, searches, internalSearch })
                                        }
                                        else chestOptionsMenu()
                                    })
                                }
                                else if (c.selection === 4) {
                                    if (slotCount) {
                                        const block = dimension.getBlock({ x, y, z })
                                        const container = block.getComponent("inventory").container

                                        for (let j = 0; j < container.size; j++) {
                                            if (container.size === container.emptySlotsCount) break
                                            const { x, y, z } = player.location
                                            let item = container.getItem(j)
                                            if (!item) continue
                                            dimension.spawnItem(item, { x, y, z })
                                            container.setItem(j, undefined)
                                        }
                                        terminalOffSound(i)
                                    }
                                    else chestOptionsMenu(true)
                                }
                                else if (c.selection === 5) {
                                    if (!remoteContr) chestOptionsMenu()
                                    else {
                                        const locateContainer = player.getTags().find(f => { return f.includes("Container #") })
                                        if (locateContainer) player.removeTag(locateContainer)
                                        player.addTag(`Container #${iChest.i + iChest.n + 1} [${[x, y, z]}]`)
                                    }
                                }
                            }
                            else getChestsItems(i, iChest, chest, dimension, page, allChestList, search, chestsPage, searches, internalSearch)
                        })
                    }
                    chestOptionsMenu()
                }
            }
            else if (r.selection > 0 && r.selection <= slotCount) {
                const chests = searches ? searches : chestsWithItems
                getItems(i, itemLocation[r.selection - 1], dimension, player, { iChest, chest, page, allChestList, search, chestsPage, searches: chests, internalSearch })
            }
            else if (r.selection === slotCount + 1) {
                if (thereIsNextPage) getChestsItems(i, 0, null, dimension, page + 1, allChestList, search, chestsPage, searches)
                else if (page) {
                    getChestsItems(i, 0, null, dimension, page - 1, allChestList, search, chestsPage, searches)
                }
                else terminal(i, search, allChestList, chestsPage, searches)
            }
            else if (r.selection === slotCount + 2) {
                if (thereIsNextPage && page) {
                    getChestsItems(i, 0, null, dimension, page - 1, allChestList, search, chestsPage, searches)
                }
                else terminal(i, search, allChestList, chestsPage, searches)
            }
            else if (r.selection === slotCount + 3) terminal(i, search, allChestList, chestsPage, searches)
        }
        else chest ? terminal(i, search, allChestList, chestsPage, searches) : moreOptions(i, search, allChestList, chestsPage, dimension, searches, 1)
    })
    function searchItems() {
        const searchMenu = new ModalFormData().title(`Search`).textField("", "Search").toggle("Cancel search")
        screenAnimation(i, 1)
        searchMenu.show(i.source).then(r => {
            if (!r.canceled) {
                if (!r.formValues[1]) {
                    getChestsItems(i, iChest, chest, dimension, page, allChestList, search, chestsPage, searches, r.formValues[0].trim().toLowerCase())
                    world.playSound("terminal.pc_pip", { x: i.block.x, y: i.block.y, z: i.block.z })
                }
                else getChestsItems(i, iChest, chest, dimension, page, allChestList, search, chestsPage, searches)
            }
            else getChestsItems(i, iChest, chest, dimension, page, allChestList, search, chestsPage, searches, internalSearch)
            screenAnimation(i)
        })
    }
}

function getItems(i, itemLocation, dimension, player, p) {
    const settings = i.source.getTags().find(f => { return f.includes('{"terminalSettings":[') })
    const [, , , , settE, ,] = settings ? JSON.parse(settings).terminalSettings : [2, 100, false, false, false, false, false]
    const { x, y, z, slot } = itemLocation
    const block = dimension.getBlock({ x, y, z })
    if (!block || !block.getComponent("inventory")) { errorMessage(); return }
    const container = block.getComponent("inventory").container
    const inventory = player.getComponent('inventory').container
    let item = container.getItem(slot)
    if (!item) { errorMessage(); return }
    const enchantments = item.hasComponent("enchantable") ? item.getComponent("enchantable").getEnchantments() : []
    const durability = item.getComponent("durability")
    const enchant = []

    const actualAmount = item.amount
    const actualTypeID = item.typeId
    const actualSize = container.size

    if (enchantments.length) {
        for (const enchantment of enchantments) {
            enchant.push(` ${enchantment.type.id.slice(0, 1).toUpperCase() + enchantment.type.id.slice(1).replace(/_/g, " ")}: ${enchantment.level}\n`)
        }
    }

    const unknownItem = `§hCould not get enough information about what type of §c${item.typeId.replace(/_/g, " ")}§h this item is.`
    const itemName = `${item.typeId.split(":")[1].slice(0, 1).toUpperCase() + item.typeId.split(":")[1].slice(1).replace(/_/g, " ")}`
    const get = new ModalFormData().title(itemName)
        .slider(`§hID: ${item.typeId}\n${durability ? `§7Durability: ${durability.maxDurability - durability.damage}/${durability.maxDurability}\n` : ""}${item.isStackable ? `§hAmount: ${item.amount}\n` : ""}\n${enchantments.length ? `§hEnchantments:§9\n${enchant.join().replace(/,/g, "")}\n` : ``}${unknownInfo.has(item.typeId) ? `${unknownItem}\n\n` : ''}§rGet`, 1, item.amount, 1).toggle(`Cancel`)

    screenAnimation(i, 1)
    get.show(player).then(r => {
        if (!r.canceled && !r.formValues[1]) {
            if (!block || !block.getComponent("inventory")) { errorMessage(); return }
            item = container.getItem(slot)
            if (!item || (actualSize != block.getComponent("inventory").container.size) || (actualAmount != item.amount) || (actualTypeID != item.typeId)) { errorMessage(item, itemName); return }
            const { x, y, z } = player.location
            const subtracted = item.amount - r.formValues[0]

            const newStack = item.clone()
            newStack.amount = r.formValues[0]

            if (inventory.emptySlotsCount) inventory.addItem(newStack)
            else dimension.spawnItem(newStack, { x, y, z })

            if (subtracted) item.amount = subtracted
            container.setItem(slot, subtracted < 1 ? undefined : item)
            if (!settE) { back(5) } else terminalOffSound(i)
        }
        else back()
        screenAnimation(i)
    })
    function errorMessage(item, itemName) {
        const messageForm = new MessageFormData().title((!block || !block.getComponent("inventory")) ? "The container no longer exists" : !item ? "The item no longer exists" : itemName)
            .body({ translate: `terminal.messageForm.container.error.text` }).button1('Close the menu').button2({ translate: `terminal.messageForm.back.chest${p.chest ? '' : 's'}.text` })
        messageForm.show(i.source).then(m => { if (!m.canceled) { if (m.selection === 1) back(); else if (m.selection === 0) terminalOffSound(i) } })
        screenAnimation(i)
    }
    function back(alerts) {
        (!block || !block.getComponent("inventory")) ? terminal(i, p.search) : getChestsItems(i, p.iChest, p.chest, dimension, p.page, p.allChestList, p.search, p.chestsPage, p.searches, p.internalSearch, alerts)
    }
}

const validContainers = [
    "minecraft:chest",
    "minecraft:trapped_chest",
    "minecraft:barrel",
    "_shulker_box"
]

async function allChests(startX, startY, startZ, dimension) {
    const chests = []
    const visited = new Set()
    let hasFinished = false

    for (const [dx, dy, dz] of [[1, 0, 0], [-1, 0, 0], [0, 0, 1], [0, 0, -1], [0, 1, 0], [0, -1, 0]]) {
        getChest(startX + dx, startY + dy, startZ + dz)
    }
    function getChest(x, y, z) {
        const cable = dimension.getBlock({ x, y, z })
        hasFinished = true

        if (cable && !visited.has(`${x},${y},${z}`) && cable.typeId === cableId) {
            const conections = connectedSides(cable)
            dimension.spawnParticle(`effect:binary`, { x: x + 0.5, y: y + 0.5, z: z + 0.5 }, new MolangVariableMap())
            visited.add(`${x},${y},${z}`)

            for (const [dx, dy, dz] of conections) {
                const newX = x + dx
                const newY = y + dy
                const newZ = z + dz

                const chestBlock = dimension.getBlock({ x: newX, y: newY, z: newZ })
                if (chestBlock && (validContainers.some(f => { return chestBlock.typeId.includes(f) })) && !visited.has(`${newX},${newY},${newZ}`)) {
                    const container = chestBlock.getComponent("inventory").container
                    const direction = chestBlock.permutation.getState("minecraft:cardinal_direction")

                    if (!chestBlock.typeId.includes("_shulker_box") && !chestBlock.typeId.includes("minecraft:barrel")) {
                        switch (direction) {
                            case "north": matchedContainers([{ xz: newX + 1 }, { xz: newX - 1 }], dimension, chestBlock, container, direction, newX, newY, newZ, true); break
                            case "south": matchedContainers([{ xz: newX + 1 }, { xz: newX - 1 }], dimension, chestBlock, container, direction, newX, newY, newZ, true); break
                            case "west": matchedContainers([{ xz: newZ + 1 }, { xz: newZ - 1 }], dimension, chestBlock, container, direction, newX, newY, newZ); break
                            case "east": matchedContainers([{ xz: newZ + 1 }, { xz: newZ - 1 }], dimension, chestBlock, container, direction, newX, newY, newZ); break
                        }
                    }
                    chests.push({ x: newX, y: newY, z: newZ })
                    visited.add(`${newX},${newY},${newZ}`)
                }
                try {
                    system.runTimeout(() => {
                        if (!visited.has(`${newX},${newY},${newZ}`)) getChest(newX, newY, newZ)
                    }, 1)
                } catch (e) { }
            }
        }
    }
    const ticks = async () => system.runTimeout(() => {
        if (ready && !hasFinished) {
            ready(chests)
            system.clearRun(returnChest)
        }
        else {
            hasFinished = false
            ticks()
        }
    }, 1)
    ticks()

    function connectedSides(cable, conections = []) {
        if (cable.permutation.getState("cable:east")) conections.push([1, 0, 0, "cable:east"])
        if (cable.permutation.getState("cable:west")) conections.push([-1, 0, 0, "cable:west"])
        if (cable.permutation.getState("cable:south")) conections.push([0, 0, 1, "cable:south"])
        if (cable.permutation.getState("cable:north")) conections.push([0, 0, -1, "cable:north"])
        if (cable.permutation.getState("cable:up")) conections.push([0, 1, 0, "cable:up"])
        if (cable.permutation.getState("cable:down")) conections.push([0, -1, 0, "cable:down"])
        return conections
    }
    function matchedContainers(sides, dimension, chestBlock, container, direction, x, y, z, isX) {
        for (const { xz } of sides) {
            const sameChest = dimension.getBlock({ x: isX ? xz : x, y: y, z: isX ? z : xz })
            if (!sameChest) visited.delete(`${x},${y},${z}`)
            else if (sameChest.typeId === chestBlock.typeId) {
                const sameContainer = sameChest.getComponent("inventory").container
                const sameDirection = sameChest.permutation.getState("minecraft:cardinal_direction")

                if (!visited.has(`${isX ? xz : x},${y},${isX ? z : xz}`) && direction === sameDirection && sameContainer.size === container.size && sameContainer.emptySlotsCount === container.emptySlotsCount) {
                    const firstSlot = container.getItem(0)
                    container.setItem(0, new ItemStack("minecraft:invisible_bedrock"))
                    if (sameContainer.getItem(0) && sameContainer.getItem(0).typeId === "minecraft:invisible_bedrock") {
                        container.setItem(0, firstSlot)
                        visited.add(`${isX ? xz : x},${y},${isX ? z : xz}`)
                        //chestsInUse.add(`${isX ? xz : x},${y},${isX ? z : xz}`)
                        break
                    }
                    container.setItem(0, firstSlot)
                }
            }
        }
    }
    let ready, returnChest;
    return new Promise((resolve, reject) => {
        ready = resolve
        returnChest = system.runTimeout(() => {
            resolve(chests); reject("Error")
        }, 2400)
    })
}

world.afterEvents.playerPlaceBlock.subscribe(b => {
    const blockId = b.block.typeId
    if (blockId === cableId || blockId === "effect99:storage_manager_terminal" || validContainers.some(f => { return blockId.includes(f) })) {
        const { x, y, z } = b.block
        b.dimension.runCommand(`fill ${x + 1} ${y + 1} ${z + 1} ${x - 1} ${y - 1} ${z - 1} ${cableId} replace ${cableId}`)
    }
})

world.afterEvents.playerBreakBlock.subscribe(b => {
    const blockId = b.brokenBlockPermutation.type.id
    if (blockId === cableId || blockId === "effect99:storage_manager_terminal" || validContainers.some(f => { return blockId.includes(f) })) {
        const { x, y, z } = b.block
        b.dimension.runCommand(`fill ${x + 1} ${y + 1} ${z + 1} ${x - 1} ${y - 1} ${z - 1} ${cableId} replace ${cableId}`)
    }
})

const unknownInfo = new Set([
    "minecraft:skull",
    "minecraft:potion",
    "minecraft:splash_potion",
    "minecraft:lingering_potion",
    "minecraft:firework_rocket",
    "minecraft:firework_star",
    "minecraft:leaves",
    "minecraft:sapling",
    "minecraft:banner",
    "minecraft:sand"
])

function terminalCommands(i, p) {
    const terminalC = new TerminalSimulator(i.source.name, world.getAllPlayers())

    function sendCommand(outData = '', pathData = `F:\\Users\\${i.source.name}\\Terminal`, command = '') {
        const tCommands = new ModalFormData().title(`Terminal`).textField(`§7${outData}\n\n §f${pathData}>`, `cmd`, command)

        if (pathData.endsWith("\\Terminal\\Options\\Inventory")) {
            for (const player of world.getAllPlayers()) {
                if (pathData.trim().endsWith(`F:\\Users\\${player.name}\\Terminal\\Options\\Inventory`)) { getInventory(i, undefined, p, player); return }
            }
        }
        tCommands.show(i.source).then(r => {
            if (!r.canceled) {
                try {
                    const { out, path } = terminalC.executeCommand(i, r.formValues[0])
                    const commd = r.formValues[0].split("/")
                    sendCommand(out, path, `${commd[0]}${commd[1] ? "/" : ""}`)
                } catch (e) { sendCommand("§cError") }
            }
            else moreOptions(i, p.search, p.loadeds, p.page, p.dimension, p.chestsWithItems, p.chestCount)
        })
    }
    sendCommand()
}